#!/usr/bin/python
# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of the Willow Garage nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Author: Wim Meeussen



from __future__ import with_statement

import roslib; roslib.load_manifest('application_manager')
from PyQt4 import QtGui, QtCore
import rospy
import actionlib
from actionlib_msgs.msg import GoalStatus
import sys
from application_msgs.msg import *


class AppButton(QtGui.QPushButton):

    def __init__(self, state, name, widget):
        QtGui.QPushButton.__init__(self, '', widget)
        self.name = name
        self.set_state(state)
        self.name_signal = QtCore.pyqtSignal()
        self.connect(self, QtCore.SIGNAL('clicked()'),
                     self.button_click)
        
    def set_state(self, state):
        self.state = state
        self.setText(self.name.split('/')[-1]+'\n'+self.state)
        

    def button_click(self):
        self.emit(QtCore.SIGNAL('clicked_name(str, str)'), self.name, self.state)


class ManagerInterface():
    def __init__(self):
        self.list = actionlib.SimpleActionClient('application_manager/list_applications', ListApplicationsAction)
        self.run = actionlib.SimpleActionClient('application_manager/run_application', RunApplicationAction)
        self.stop = actionlib.SimpleActionClient('application_manager/stop_application', StopApplicationAction)
        self.kill = actionlib.SimpleActionClient('application_manager/kill_application', KillApplicationAction)

        self.list.wait_for_server()
        self.run.wait_for_server()
        self.stop.wait_for_server()
        self.kill.wait_for_server()


    def change_state(self, application, state):
        if state == 'Stopped':
            self.run.send_goal(RunApplicationGoal(application, True))
            print 'Running %s'%application
        elif state == 'Running' or state == 'Starting':
            self.stop.send_goal_and_wait(StopApplicationGoal(application))
            print 'Stopping %s'%application
        elif state == 'Preempting':
            self.kill.send_goal_and_wait(KillApplicationGoal(application))            
            print 'Killing %s'%application
    

    def get_state(self):
        self.list.send_goal_and_wait(ListApplicationsGoal())
        res = self.list.get_result()
        return zip(res.application_name, res.application_status)

        

class AppManagerGui(QtGui.QMainWindow):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.manager = ManagerInterface()

        self.setWindowTitle('Application Manager Interface')
        self.center()

        self.buttons = {}
        self.update_button_state()


        self.timer = QtCore.QTimer(self)
        self.timer.connect(self.timer, QtCore.SIGNAL('timeout()'), self.update_button_state)
        self.timer.start(1000)


    def button_click(self, name, state):
        self.manager.change_state(name, state)
        self.update_button_state()


    def update_button_state(self):
        new_buttons = {}
        offset = 10
        for name, state in self.manager.get_state():
            # create button list
            if not name in self.buttons:
                button = AppButton(state, name, self)
                button.connect(button, QtCore.SIGNAL('clicked_name(str, str)'), self.button_click)
            else:
                button = self.buttons[name]
                button.set_state(state)
            button.setGeometry(10, offset, 160, 70)
            new_buttons[name] = button
            offset = offset+80
        self.buttons = new_buttons
        self.resize(180,len(self.buttons)*80+10)

    def center(self):
        screen = QtGui.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move((screen.width()-size.width())/2, (screen.height()-size.height())/2)


def main():
    app = QtGui.QApplication(sys.argv)
    app_manager_gui = AppManagerGui()
    app_manager_gui.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
   rospy.init_node('application_manager_do')
   main()



